<?
/**
 * @package rbcore
 * File:  db.php
 * Database utility functions
 * Author:  Robocommerce, LLC (Zvi Boshernitzan and Brian Rosenthal)
 * Copyright 2004, 2005:  Robocommerce, LLC (http://www.robocommerce.com)
 * Distributed under Gnu General Public License (see includes/gpl_general.txt)
 * Version 1.0
 */


/**
 * functions for database queries.
 * 
 * depends on {@link rba::to_obj()}, {@link rb::elapsed()}, {@link rbok::identifier()}
 * - query functions have been refactored to
 *
 *   - db::object    [s](  {$sql | $tbl,$where}  ,  $class_name='generic'  );
 *
 *   - db::{dict|arr}[s](  {$sql | $tbl,$where}  	  	);  // mysql_fetch_array
 *   - db::assoc     [s](  {$sql | $tbl,$where}  	  	);  // mysql_fetch_assoc
 *   - db::row       [s](  {$sql | $tbl,$where}  	  	);  // mysql_fetch_row
 *
 *   - db::string    [s](  {$sql | $tbl,$field,$where}  );  // single value per row
 *
 *   - db::sum          (          $tbl,$field,$where  	);  // sum of single value per row
 *   - db::exists       (          $tbl,$where         	);  // count of single value per row
 *   - db::count        (          $tbl,$where         	);  // same as exists
 * 
 * @package rbcore
 */
class db {

    function conf($key = null, $val = null) {
        static $confs = array('debug' => false);
        if (is_null($key)) return $confs;
        if (is_null($val)) return $confs[$key];
        $old_val = $confs[$key];
        $confs[$key] = $val;
        return $old_val;
    }
	
	////////////////////////////////
	// FUNCTIONS THAT FORMAT TEXT //
	////////////////////////////////

	/**
	 * escapes the value without putting quotes arround it.
	 *
	 */
	function escape($s) {
		if (is_string($s)) {
			return trim(stripslashes($s));
		} elseif (is_array($s)) {
			reset($s);
			while (list($key, $value) = each($s)) {
				$s[$key] = db::escape($value);
			}
			return $s;
		} else {
			return $s;
		}

	}
	/**
	 * escapes an arbitrary value for placement into a sql statement.
	 *
	 * may be tested with {@link db::test_sql_escape()}.
	 */
	function sql_escape($s) {
		if (gettype($s) == "boolean")	 	{$s = ($s ? "1" : "0");}
		elseif (is_numeric($s))		   		{$s = "'" . db::sql_clean_string($s) . "'";}
		elseif (is_null($s))			  	{$s = "null";}
		elseif ($s == "null")			 	{$s = "null";}
		elseif ($s == "now()")				{$s = "now()";}
		elseif (is_array($s) and !empty($s)){$s = "('" . join("','",db::escape($s)) . "')";}
		else							  	{$s = "'" . db::sql_clean_string($s) . "'";}
		return $s;
	}

	function sql_clean_string($s) { return mysql_escape_string(stripslashes($s)); }

	function sql_safe_params($dict_or_string) {
		if (rbok::id($dict_or_string)) return db::sql_params(array('id' => $dict_or_string));
		if (is_array($dict_or_string)) return db::sql_params($dict_or_string);
		else return $dict_or_string;
	}

	function where($dict) {return db::sql_params($dict);}
 	function where_clause($dict) {return db::sql_params($dict);}

	function sql_params($dict) {
		assert('is_array($dict)');
		assert('count($dict) > 0');
		$parts = array();
		foreach($dict as $key => $value) {
			// if key is not identifier, then do not add '=' to it.
			// if value is an array, then use 'in', perhaps.
			if (!is_array($value)) {
				$joiner = ' '.(is_null($value)?'is':'=').' ';
				if (!rbok::sql_identifier($key)) $joiner = ' ';
				$parts[] = $key . $joiner . sql_escape($value);
			} elseif (!empty($value)) {
				// we have a non-empty array
				$joiner = ' in ';
				if (!rbok::sql_identifier($key)) $joiner = ' ';
				$parts[] = $key . $joiner . sql_escape($value);
			}
		}
		
		$vals = join("\n   and ",$parts);
		return $vals;
	}



	/////////////////////////////////
	// FUNCTIONS THAT GENERATE SQL //
	///////////////////////////////// 

	function sql_insert($table,$dict) {
		foreach ($dict as $key => $value) {
			$keys[] = $key;
			$values[] = sql_escape($value);
		}
		$cols = join(",\n  ",$keys);
		$vals = join(",\n  ",$values);
		return "
			insert into $table (
					$cols
					) values (
						$vals
						)";
	}

	function sql_update($table,$dict,$where='') {
		$where = db::sql_safe_params($where);
		if (!is_array($dict)) {
			$error_msg = "sql_update requires dictionary: ";
			echo_r(compact('error_msg','table','dict','where'));
		}
		foreach($dict as $key => $value) {
			$parts[] = $key . " = " . sql_escape($value);
		}
		$vals = join(",\n  ",$parts);
		return "
			update $table set
			$vals
			where $where";
	}

	////////////////////////////////////////
	// FUNCTIONS THAT MAKE DATABASE CALLS //
	////////////////////////////////////////
	
	// ------------- basic query functions --------------
	/**
	 * resolves tbl/where or sql into sql, executes it, and returns the recordset.
	 */
	function _args($args) {
		if (rbok::identifier($args[0])) {
			list($tbl,$where) = $args;
			$where = db::sql_safe_params($where);
			$sql = "select * from " . $tbl . " where " . $where ;
		} else {
			list($sql) = $args;
		}
		return db::q($sql);
	}
	function dict() {$rs = db::_args(func_get_args()); return db::fetch_array($rs);}
	function dicts() {
		$rs = db::_args(func_get_args());
		while ($r = db::fetch_array($rs)) $a[] = $r;
		return coalesce($a,array());
	}
	function arr() {$rs = db::_args(func_get_args()); return db::fetch_array($rs);}
	function arrs() {
		$rs = db::_args(func_get_args());
		while ($r = db::fetch_array($rs)) $a[] = $r;
		return coalesce($a,array());
	}
	function assoc() {$rs = db::_args(func_get_args()); return db::fetch_assoc($rs);}
	function assocs() {
		$rs = db::_args(func_get_args());
		while ($r = db::fetch_assoc($rs)) $a[] = $r;
		return coalesce($a,array());
	}
	function row() {$rs = db::_args(func_get_args()); return db::fetch_row($rs);}
	function rows() {
		$rs = db::_args(func_get_args());
		while ($r = db::fetch_row($rs)) $a[] = $r;
		return coalesce($a,array());
	}

	function nextid($tbl,$col,$val) {
		return (int)db::string(sprintf('select min(%s) from %s where %s > %s',$col,$tbl,$col,$val));
	}
	function previd($tbl,$col,$val) {
		return (int)db::string(sprintf('select max(%s) from %s where %s < %s',$col,$tbl,$col,$val));
	}
	
	// ------------- class-based functions
	/**
	 * resolves tbl/where/classname or sql/classname into sql/classname (with 'generic' default)
	 *  and executes the sql.
	 */
	function _args_cls($args) {
		if (rbok::identifier($args[0])) {
			list($tbl,$where,$classname) = $args;
			$where = db::sql_safe_params($where);
			$sql = "select * from " . $tbl . " where " . $where ;
		} else {
			list($sql,$classname) = $args;
		}
		$classname = coalesce($classname,'generic');
		return array(db::q($sql),$classname);
	}
	function object() {
		list($rs,$cls) = db::_args_cls(func_get_args());
		return rba::to_obj(db::fetch_array($rs), $cls);
	}
	function objects() {
		list($rs,$cls) = db::_args_cls(func_get_args());
		while ($r = db::fetch_array($rs)) $a[] = rba::to_obj($r, $cls);
		return coalesce($a,array());
	}
	
	function kvdict() {
		list($rs,$cls) = db::_args_cls(func_get_args());
		while ($r = db::fetch_row($rs)) {
			$a[$r[0]] = $r[1];
		}
		return coalesce($a,array());
	}

	// ------------ field-based functions
	/**
	 * resolves tbl/field/where or sql into sql, executes it, and returns the recordset.
	 */
	function _args_fld($args) {
		if (rbok::identifier($args[0])) {
			list($tbl,$field,$where) = $args;
			$where = db::sql_safe_params($where);
			$sql = 'select ' . $field . ' from ' . $tbl . ' where ' . $where ;
		} else {
			$sql = $args[0];
		}
		return db::q($sql);
	}
	
	function string() {
		$rs = db::_args_fld(func_get_args());
		if ($r = db::fetch_row($rs)) return $r[0];
		return '';
	}
	function strings() {
		$rs = db::_args_fld(func_get_args());
		while($r = db::fetch_row($rs)) $a[] = $r[0];
		return coalesce($a,array());
	}
	// -- non-polymorphic functions
	function sum($table_name, $field, $where) {
		$where = db::sql_safe_params($where);
		$sql = ("select sum(" . $field . ") from " . $table_name 
				. (empty($where) ? '' : ' where ' . $where)
				);
		return db::string($sql);		
	}	
	function max($table_name, $field, $where) {
		$where = db::sql_safe_params($where);
		$sql = ("select coalesce(max(" . $field . "), 0) from " . $table_name 
				. (empty($where) ? '' : ' where ' . $where)
				);
		return db::string($sql);		
	}
	/**
	 * checks to see if a row exists in a table.
	 * 
	 * also accepts a sql query, which gets stripped down to the counting essentials.
	 */
	function exists($table, $where = null)		{ return db::count($table, $where); }
	function count($table, $where = null) {
		if (rbok::sql_identifier($table)) {
			$where = db::sql_safe_params($where);
			$q = "select count(1) from " . $table . " where " . $where;
		} else {
			$q = preg_replace("/^.*from/","select count(1) as c from",$table);
			$q = preg_replace("/limit.*$/","",$q);
		}
		return (int)db::string($q);
	}
	
	// ---------------- older functions

	
	// multirow
	/**
	 * @deprecated
	 */
	function db_to_object_array($table_name,$where_clause,$class_name='generic') {
		$where_clause = db::sql_safe_params($where_clause);
		$q = "select * from " . $table_name . " where " . $where_clause ;
		return db::sql_to_object_array($q, $class_name);
	}
	/**
	 * @deprecated
	 */
	function sql_to_array($q) {
		$rs = db::q($q);
		$arr = array();
		while ($record = db::fetch_array($rs)) {
			$arr[] = $record;
		}
		return $arr;
	}
	/**
	 * @deprecated
	 */
	function sql_to_row($q)			{ return db::row($q); }
	
	/**
	 * @deprecated
	 */
	function sql_to_object_array($q, $class_name = "generic") {
		$rs = db::q($q);
		$arr = array();
		while ($record = db::fetch_array($rs)) {
			$arr[] = rba::to_obj($record, $class_name);
		}
		return $arr;
	}
	/**
	 * @deprecated
	 */
	function sql_to_object($q,$class_name = 'generic') {
		$arr = sql_to_object_array($q, $class_name = 'generic');
		if (count($arr) > 0) {
			$obj = $arr[0];
			return $obj;
		}
		$retval = db::string($q);
		return $retval;
	}
	/**
	 * @deprecated
	 */
	function db_to_object($table_name,$where_clause,$class_name='generic') {
		$where_clause = db::sql_safe_params($where_clause);
		$q = "select * from " . $table_name . " where " . $where_clause ;
		return db::sql_to_object($q, $class_name);
	}



	// This function is used by referrer_benefit_management
	/**
	 * why!?
	 */
	function select($q, $debug = false) {
		$rows = db::sql_to_array($q);
		$arr = array();
		foreach ($rows as $row) {
			foreach ($row as $key => $value) {
				$arr[] = $value;
			}
		}
		return $arr;
	}


	// -- data modification statements (delete/insert/update)
	
	/**
	 * @todo support multi-insert statements
	 */
	function insert($table, $dict, $debug = false) {
		$q = db::sql_insert($table,$dict);
		db::db_exec($q, "db_link", $debug);
		$id = db::last_insert_id();
		return $id;
	}
	function update($table,$dict,$where='', $debug = false) {
		$where = db::sql_safe_params($where);
		$q = db::sql_update($table,$dict,$where);
		return db::db_exec($q, "db_link", $debug);
	}
	function update_or_insert($table,$dict,$where='',$debug = false) {
		$where_string = db::sql_safe_params($where);
		if (db::exists($table,$where_string)) {
			return db::update($table,$dict,$where_string,$debug);
		} else {
			if (is_array($where)) {
				return db::insert($table,array_merge($dict, $where),$debug);
			}
			else {
				return db::insert($table,$dict,$debug);
			}
		}
	}
	function delete($table, $where, $debug = false) {
		$where = db::sql_safe_params($where);
		$q = "delete from " . $table . " where " . $where . ";";
		db::db_exec($q, "db_link", $debug);
		assert('!db::exists($table,$where)');
	}
	/**
	 * recursively deletes rows from multiple tables following key relations.
	 */
	function cascade_delete($table_name,$column_name,$ids) {
		// spec = array(table_name,column_name,([parent_column_name]|ids),[child_spec],[child_spec],...)
		$specs = array_slice(func_get_args(),3);
		if (is_array($ids)) {
			for ($i = 0; $i < count($ids); $i++) {
				if (!is_numeric($ids[$i])) {
					$ids[$i] = sql_escape($ids[$i]);
				}
			}
		}
		$where_clause = $column_name . " in (" . join(',',$ids) . ")";
		if (count($specs)) {
			$query_cols = array();
			foreach ($specs as $aspec) {
				$parent_column = $aspec[2];
				if (empty($parent_column)) $parent_column = $aspec[1];
				$query_cols[$parent_column][] = $aspec;
			}
			$cols = array_keys($query_cols);
			$vals = db::sql_to_array("select " . join(",",$cols) . " from " . $table_name . " where " . $where_clause);
			$val_ids = array();
			// UNIQUIFY IDS BY COLUMN
			foreach($vals as $row) {
				foreach($cols as $col) {$val_ids[$col][$row[$col]] = 1;}
			}
			foreach($query_cols as $col => $specs) {
				if (count($val_ids[$col]) > 0) {
					$new_ids = array_keys($val_ids[$col]);
					if (!empty($new_ids) and (count($new_ids) > 0)) {
						foreach($specs as $spec) {
							// replace the parent_column_name with the new_ids.
							$spec[2] = $new_ids;
							call_user_func_array(array('db', 'cascade_delete'),$spec);
						}
					}
				}
			}
		} // endif !empty($children_spec)
		// delete the parent specification
		if (!empty($ids) and (count($ids) > 0)) {
			db::delete($table_name,$where_clause);
		}
	}

	/**
	 *
	 * @deprecated
	 */
	function perform($table, $data, $action = 'insert', $where = '', $link = 'db_link') {
		$where = db::sql_safe_params($where);
		reset($data);
		if ($action == 'insert') {
			$query = db::sql_insert($table,$data);
		} elseif ($action == 'update') {
			$query = db::sql_update($table,$data,$where);
		}
		return db::db_exec($query, $link);
	}

	
	
	function query($q, $link = 'db_link', $debug = false) { return db::db_exec($q, $link, $debug); }
	function q($q, $link = 'db_link', $debug = false) { return db::db_exec($q, $link, $debug); }
	function db_exec($query, $link = 'db_link', $debug = false) {
		static $n_queries = 0;
		
		// if it is a resultset, just pass it through.
		if (is_resource($query)) return $query;
		
		if (empty($link)) $link = 'db_link';
		if (defined('STORE_DB_TRANSACTIONS') && (STORE_DB_TRANSACTIONS == 'true')) {
			; //error_log('QUERY ' . $query . "\n", 3, STORE_PAGE_PARSE_TIME_LOG);
		}
		// echo_r(debug_backtrace());
		$debug = db::conf('debug');
		if ($debug) {
			$n_queries += 1;
			rb::elapsed('reset');
			flush();
		}
	
		$actually_run = true;
		if ($debug) {
		#	$actually_run = false;
			if (!preg_match("/delete|insert|update/i",$query)) $actually_run = true;
		}
		if ($actually_run) {
			$result = mysql_query($query, $GLOBALS[$link]) or db::error($query, mysql_errno(), mysql_error());
			if ($debug) {
				rb::elapsed(' [' . $n_queries . ']' . $query);
				flush();
			}
			
			if (defined('STORE_DB_TRANSACTIONS') && (STORE_DB_TRANSACTIONS == 'true')) {
				$result_error = mysql_error();
				//error_log('RESULT ' . $result . ' ' . $result_error . "\n", 3, STORE_PAGE_PARSE_TIME_LOG);
			}

			return $result;
		}
	}


	///////////////////////////
	// CONNECTION MANAGEMENT //
	///////////////////////////

	function connect($server = DB_SERVER
			, $username = DB_SERVER_USERNAME
			, $password = DB_SERVER_PASSWORD
			, $database = DB_DATABASE
			, $link = 'db_link') {
		if (empty($link)) $link = 'db_link';
		
		$GLOBALS[$link] = mysql_connect($server,$username,$password);
		if ($GLOBALS[$link]) mysql_select_db($database);

		return $GLOBALS[$link];
	}

	function close($link = 'db_link') {
		return mysql_close($GLOBALS[$link]);
	}

	function error($query, $errno, $error) {
		global $HTTP_REFERER;
		echo_r(compact('error', 'query', 'errno'));
		exit();
	}

	////////////////////////////////
	// DATABASE WRAPPER FUNCTIONS //
	////////////////////////////////

	function fetch_array($db_query) { return mysql_fetch_array($db_query, MYSQL_ASSOC); }
	function fetch_assoc($db_query) { return mysql_fetch_assoc($db_query); }
	function fetch_row($db_query)   { return mysql_fetch_row($db_query); }
	function num_rows($db_query) { return mysql_num_rows($db_query); }
	function data_seek($db_query, $row_number) { return mysql_data_seek($db_query, $row_number); }
	function last_insert_id() { return mysql_insert_id(); }
	function free_result($db_query) { return mysql_free_result($db_query); }
	function fetch_fields($db_query) { return mysql_fetch_field($db_query); }
	function output($string) { return htmlspecialchars($string); }
	function input($string) { return addslashes($string); }


}




?>
